@@ -1102,14 +1102,15 @@ void pyi_win32_minimize_console()
 \**********************************************************************/
 /* Our last resort in ensuring that onefile application can clean up
  * its temporary directory... */
-void
-pyi_win32_force_unload_bundled_dlls(struct PYI_CONTEXT *pyi_ctx)
+static int
+_pyi_win32_force_unload_bundled_dlls(const struct PYI_CONTEXT *pyi_ctx)
 {
     HANDLE process_handle;
     HMODULE *loaded_dlls = NULL;
     DWORD loaded_dlls_size = 0;
     int num_dlls;
     int num_problematic_dlls;
+    int num_unloaded_dlls = 0;
     wchar_t dll_filename[PYI_PATH_MAX];
     wchar_t application_home_dir[PYI_PATH_MAX];
     size_t application_home_dir_len;
@@ -1119,16 +1120,16 @@ pyi_win32_force_unload_bundled_dlls(struct PYI_CONTEXT *pyi_ctx)
 
     /* Query the required size for lpModules */
     if (EnumProcessModules(process_handle, loaded_dlls, 0, &loaded_dlls_size) == 0) {
-        return;
+        return 0; /* Technically an error, but treat it as if no DLLs were unloaded. */
     }
     if (loaded_dlls_size <= 0) {
-        return;
+        return 0;
     }
 
     /* Allocate the array */
     loaded_dlls = calloc(loaded_dlls_size, 1);
     if (!loaded_dlls) {
-        return;
+        return 0;
     }
 
     /* Read the loaded DLLs handles into array */
@@ -1180,10 +1181,11 @@ pyi_win32_force_unload_bundled_dlls(struct PYI_CONTEXT *pyi_ctx)
         /* Keep calling FreeLibrary() until it fails - which hopefully
          * means that the offending DLL is gone for good. */
         while (1) {
-            PYI_DEBUG_W(L"LOADER: forcing unload of %ls\n", dll_filename);
             num_attempts++;
+            PYI_DEBUG_W(L"LOADER: forcing unload of %ls (attempt #%d)\n", dll_filename, num_attempts);
             if (FreeLibrary(loaded_dlls[i]) == 0) {
                 PYI_DEBUG_W(L"LOADER: DLL unloaded after %d attempt(s)!\n", num_attempts);
+                num_unloaded_dlls++;
                 break;
             }
             /* Make sure we don't loop forever, just in case. */
@@ -1196,6 +1198,83 @@ pyi_win32_force_unload_bundled_dlls(struct PYI_CONTEXT *pyi_ctx)
 
 cleanup:
     free(loaded_dlls);
+
+    return num_unloaded_dlls;
+}
+
+int pyi_win32_mitigate_locked_temporary_directory(const struct PYI_CONTEXT *pyi_ctx)
+{
+    int num_unloaded_dlls;
+    int cleanup_status = -1;
+
+    const int max_attempts = 15; /* Maximum number of attempts in retry loop */
+    const int delay = 1000; /* Delay for retry loop (milliseconds) */
+    int attempt_count;
+
+    /* One reason for locked files in the temporary directory might be
+     * due to Tcl/Tk shared libs pulling in dependencies and failing to
+     * release them when they are unloaded. This happens in frozen
+     * applications that use splash screen, where Tcl/Tk needs to be
+     * loaded in the parent process of the onefile application.
+     *
+     * For example, tcl86.dll from mingw-w64-i686-tcl 8.6.12-3 (32-bit
+     * msys2/mingw32 environment) pulls in libgcc_s_dw2-1.dll and
+     * libwinpthread-1.dll, and does not unload them when it is unloaded.
+     * Similar situation was observed in 64-bit builds with UPX-processed
+     * Tcl/Tk DLLs, which leak VCRUNTIME140.dll (although this is mitigated
+     * separately by pre-load of system copy, if available) and/or ZLIB1.dll.
+     *
+     * So we go over DLLs loaded in the process, find the ones that
+     * originate from the application's temporary directory, and try
+     * to force-unload them, before repeating the directory removal
+     * attempt. Force-unloading DLLs is risky and might crash the process,
+     * but at this point, we have nothing left to lose... */
+    PYI_DEBUG_W(L"LOADER: trying to force-unload bundled DLLs from this process...\n");
+    num_unloaded_dlls = _pyi_win32_force_unload_bundled_dlls(pyi_ctx);
+    if (num_unloaded_dlls > 0) {
+        PYI_DEBUG_W(L"LOADER: unloaded %d bundled DLL(s) from this process - trying to remove temporary directory again...\n", num_unloaded_dlls);
+        cleanup_status = pyi_recursive_rmdir(pyi_ctx->application_home_dir);
+        if (cleanup_status == 0) {
+            PYI_DEBUG_W(L"LOADER: removal succeeded.\n");
+            return 0;
+        } else {
+            PYI_DEBUG_W(L"LOADER: removal failed!\n");
+        }
+    } else {
+        PYI_DEBUG_W(L"LOADER: no bundled DLLs were unloaded from this process.\n");
+    }
+
+    /* Sometimes, the spawned grandchild processes need a bit more time to
+     * completely finish; if their executable is part of the frozen
+     * application, it and its bundled dependencies will still be locked
+     * at this point. This seems to happen, for example, with the
+     * QtWebEngineProcess.exe helper when running a QtWebEngine-based
+     * application on a system with high CPU load.
+     *
+     * Other times, it might be Windows itself getting in our way; it seems
+     * that when running an x86_64 frozen application on an arm64 system,
+     * the OS ends up locking (some of?) the python extensions and DLLs
+     * for a short while; see #8837. The same issue has also been observed
+     * on x86_64; see #8866.
+     *
+     * The retry loop below attempts to address this. */
+    attempt_count = 0;
+    while (attempt_count++ < max_attempts) {
+        PYI_DEBUG_W(L"LOADER: waiting %d milliseconds before trying to remove temporary directory again...\n", delay);
+        Sleep(delay);
+        PYI_DEBUG_W(L"LOADER: trying to remove temporary directory (attempt %d / %d)...\n", attempt_count, max_attempts);
+        cleanup_status = pyi_recursive_rmdir(pyi_ctx->application_home_dir);
+        if (cleanup_status == 0) {
+            PYI_DEBUG_W(L"LOADER: removal succeeded.\n");
+            return 0;
+        } else {
+            PYI_DEBUG_W(L"LOADER: removal failed!\n");
+        }
+    }
+    PYI_DEBUG_W(L"LOADER: given up after %d attempts!\n", max_attempts);
+
+    /* Well, tough luck. */
+    return -1;
 }
 
 
